@@ -1,4 +1,7 @@
+(require 'magit)
+
 (require 'assoc)
+(eval-when-compile (require 'cl))
 
 (defvar magit-key-mode-key-maps '()
   "This will be filled lazily with proper `define-key' built
@@ -20,10 +23,10 @@
   '((logging
      (man-page "git-log")
      (actions
-      ("l" "Short" magit-display-log)
+      ("l" "Short" magit-log)
       ("L" "Long" magit-log-long)
       ("h" "Reflog" magit-reflog)
-      ("rl" "Ranged short" magit-display-log-ranged)
+      ("rl" "Ranged short" magit-log-ranged)
       ("rL" "Ranged long" magit-log-long-ranged)
       ("rh" "Ranged reflog" magit-reflog-ranged))
      (switches
@@ -55,7 +58,9 @@
      (actions
       ("f" "Current" magit-fetch-current)
       ("a" "All" magit-remote-update)
-      ("o" "Other" magit-fetch)))
+      ("o" "Other" magit-fetch))
+     (switches
+      ("-p" "Prune" "--prune")))
 
     (pushing
      (man-page "git-push")
@@ -64,7 +69,8 @@
       ("t" "Push tags" magit-push-tags))
      (switches
       ("-f" "Force" "--force")
-      ("-d" "Dry run" "-n")))
+      ("-d" "Dry run" "-n")
+      ("-u" "Set upstream" "-u")))
 
     (pulling
      (man-page "git-pull")
@@ -80,6 +86,7 @@
       ("n" "New" magit-create-branch)
       ("m" "Move" magit-move-branch)
       ("d" "Delete" magit-delete-branch)
+      ("D" "Force Delete" magit-delete-branch-forced)
       ("b" "Checkout" magit-checkout)))
 
     (tagging
@@ -108,7 +115,7 @@
       ("-nc" "No commit" "--no-commit")
       ("-sq" "Squash" "--squash"))
      (arguments
-      ("-st" "Strategy" "--strategy" read-from-minibuffer)))
+      ("-st" "Strategy" "--strategy=" read-from-minibuffer)))
 
     (rewriting
      (actions
@@ -125,7 +132,19 @@
       ("u" "Update" magit-submodule-update)
       ("b" "Both update and init" magit-submodule-update-init)
       ("i" "Init" magit-submodule-init)
-      ("s" "Sync" magit-submodule-sync))))
+      ("s" "Sync" magit-submodule-sync)))
+
+    (bisecting
+     (man-page "git-bisect")
+     (actions
+      ("b" "Bad" magit-bisect-bad)
+      ("g" "Good" magit-bisect-good)
+      ("k" "Skip" magit-bisect-skip)
+      ("l" "Log" magit-bisect-log)
+      ("r" "Reset" magit-bisect-reset)
+      ("s" "Start" magit-bisect-start)
+      ("u" "Run" magit-bisect-run)
+      ("v" "Visualize" magit-bisect-visualize))))
   "Holds the key, help, function mapping for the log-mode. If you
   modify this make sure you reset `magit-key-mode-key-maps' to
   nil.")
@@ -201,15 +220,22 @@ the group FOR-GROUP."
   "Provide help for a key (which the user is prompted for) within
 FOR-GROUP."
   (let* ((opts (magit-key-mode-options-for-group for-group))
-         (seq (read-key-sequence "Enter command prefix: "))
+         (man-page (cadr (assoc 'man-page opts)))
+         (seq (read-key-sequence
+               (format "Enter command prefix%s: "
+                       (if man-page
+                         (format ", `?' for man `%s'" man-page)
+                         ""))))
          (actions (cdr (assoc 'actions opts))))
-    ;; is it an action? If so popup the help for the to-be-run
-    ;; function
-    (if (assoc seq actions)
-        (describe-function (nth 2 (assoc seq actions)))
-      ;; otherwise give the user a man page
-      (man (or (cadr (assoc 'man-page opts))
-               (error "No help associated with %s" seq))))))
+    (cond
+      ;; if it is an action popup the help for the to-be-run function
+      ((assoc seq actions) (describe-function (nth 2 (assoc seq actions))))
+      ;; if there is "?" show a man page if there is one
+      ((equal seq "?")
+       (if man-page
+         (man man-page)
+         (error "No man page associated with `%s'" for-group)))
+      (t (error "No help associated with `%s'" seq)))))
 
 (defun magit-key-mode-exec-at-point ()
   "Run action/args/option at point."
@@ -225,42 +251,51 @@ put it in magit-key-mode-key-maps for fast lookup."
   (let* ((options (magit-key-mode-options-for-group for-group))
          (actions (cdr (assoc 'actions options)))
          (switches (cdr (assoc 'switches options)))
-         (arguments (cdr (assoc 'arguments options))))
-    (let ((map (make-sparse-keymap)))
-      ;; ret dwim
-      (define-key map (kbd "RET") 'magit-key-mode-exec-at-point)
+         (arguments (cdr (assoc 'arguments options)))
+         (map (make-sparse-keymap)))
+    (suppress-keymap map 'nodigits)
+    ;; ret dwim
+    (define-key map (kbd "RET") 'magit-key-mode-exec-at-point)
 
-      ;; all maps should 'quit' with C-g
-      (define-key map (kbd "C-g") (lambda ()
-                                    (interactive)
-                                    (magit-key-mode-command nil)))
-      ;; run help
-      (define-key map (kbd "?") `(lambda ()
-                                  (interactive)
-                                  (magit-key-mode-help ',for-group)))
+    ;; all maps should `quit' with `C-g' or `q'
+    (define-key map (kbd "C-g") `(lambda ()
+                                   (interactive)
+                                   (magit-key-mode-command nil)))
+    (define-key map (kbd "q")   `(lambda ()
+                                   (interactive)
+                                   (magit-key-mode-command nil)))
+    ;; run help
+    (define-key map (kbd "?") `(lambda ()
+                                 (interactive)
+                                 (magit-key-mode-help ',for-group)))
 
+    (flet ((defkey (k action)
+             (when (and (lookup-key map (car k))
+                        (not (numberp (lookup-key map (car k)))))
+               (message "Warning: overriding binding for `%s' in %S"
+                        (car k) for-group)
+               (ding)
+               (sit-for 2))
+             (define-key map (car k)
+               `(lambda () (interactive) ,action))))
       (when actions
         (dolist (k actions)
-          (define-key map (car k) `(lambda ()
-                                     (interactive)
-                                     (magit-key-mode-command ',(nth 2 k))))))
+          (defkey k `(magit-key-mode-command ',(nth 2 k)))))
       (when switches
         (dolist (k switches)
-          (define-key map (car k) `(lambda ()
-                                     (interactive)
-                                     (magit-key-mode-add-option
-                                      ',for-group
-                                      ,(nth 2 k))))))
+          (defkey k `(magit-key-mode-add-option ',for-group ,(nth 2 k)))))
       (when arguments
         (dolist (k arguments)
-          (define-key map (car k) `(lambda ()
-                                     (interactive)
-                                     (magit-key-mode-add-argument
-                                      ',for-group
-                                      ,(nth 2 k)
-                                      ',(nth 3 k))))))
-      (aput 'magit-key-mode-key-maps for-group map)
-      map)))
+          (defkey k `(magit-key-mode-add-argument
+                      ',for-group ,(nth 2 k) ',(nth 3 k))))))
+
+    (aput 'magit-key-mode-key-maps for-group map)
+    map))
+
+(defvar magit-key-mode-prefix nil
+  "For internal use.  Holds the prefix argument to the command
+that brought up the key-mode window, so it can be used by the
+command that's eventually invoked.")
 
 (defun magit-key-mode-command (func)
   (let ((args '()))
@@ -268,7 +303,8 @@ put it in magit-key-mode-key-maps for fast lookup."
     (maphash (lambda (k v)
                (push (concat k (shell-quote-argument v)) args))
              magit-key-mode-current-args)
-    (let ((magit-custom-options (append args magit-key-mode-current-options)))
+    (let ((magit-custom-options (append args magit-key-mode-current-options))
+          (current-prefix-arg (or current-prefix-arg magit-key-mode-prefix)))
       (set-window-configuration magit-log-mode-window-conf)
       (when func
         (call-interactively func))
@@ -325,6 +361,7 @@ highlighted before the description."
     (set (make-local-variable
           'magit-key-mode-current-args)
          (make-hash-table))
+    (set (make-local-variable 'magit-key-mode-prefix) current-prefix-arg)
     (magit-key-mode-redraw for-group))
   (message
    (concat
